from logging import getLogger

import numpy as np
from itertools import product

from typing import Tuple

__all__ = [
    'extract_simple_patches_2d',
    'reconstruct_from_simple_patches_2d',
]

logger = getLogger(__name__)


def extract_simple_patches_2d(image: np.ndarray, patch_size: Tuple[int, int]) -> np.ndarray:
    """Reshape a 2D image into a collection of patches without duplication of extracted range.
    """

    i_h, i_w = image.shape[:2]
    p_h, p_w = patch_size

    if i_h % p_h != 0 or i_w % p_w != 0:
        logger.warning(
            'image %s divided by patch %s is not zero and some parts will be lost', image.shape[:2], patch_size)

    image = image.reshape((i_h, i_w, -1))
    n_colors = image.shape[-1]

    patches = []

    n_h = int(i_h / p_h)
    n_w = int(i_w / p_w)

    for i in range(n_h):
        for j in range(n_w):
            patch = image[p_h * i:p_h * i + p_h, p_w * j:p_w * j + p_w]
            patches.append(patch.flatten())

    n_patches = len(patches)

    patches_ret = np.asarray(patches).flatten().reshape(-1, p_h, p_w, n_colors)
    if patches_ret.shape[-1] == 1:
        return patches_ret.reshape((n_patches, p_h, p_w))
    else:
        return patches_ret


def reconstruct_from_simple_patches_2d(patches: np.ndarray, image_size: Tuple[int, int]) -> np.ndarray:
    """Reconstruct the image from all of its patches.
    """
    i_h, i_w = image_size[:2]
    p_h, p_w = patches.shape[1:3]
    image = np.zeros(image_size)

    n_h = int(i_h / p_h)
    n_w = int(i_w / p_w)
    for p, (i, j) in zip(patches, product(range(n_h), range(n_w))):
        image[p_h * i:p_h * i + p_h, p_w * j:p_w * j + p_w] += p
    return image



from setuptools import setup, find_packages

setup(
    name='spm-image',
    author='Takashi Someda',
    author_email='takashi@hacarus.com',
    url='https://github.com/hacarus/spm-image',
    description='Sparse modeling and Compressive sensing in Python',
    version='0.0.2',
    packages=find_packages(),
    test_suite='tests'
)



from unittest import TestCase

from spmimage.feature_extraction.image import extract_simple_patches_2d, reconstruct_from_simple_patches_2d

import numpy as np


class TestImage(TestCase):
    def test_extract_simple_patches_2d_gray(self):
        image = np.arange(16).reshape((4, 4))

        actual = extract_simple_patches_2d(image, (2, 2,))
        self.assertEqual((4, 2, 2), actual.shape)

        self.assertTrue((np.array([0, 1, 4, 5]) == actual[0].flatten()).all())
        self.assertTrue((np.array([2, 3, 6, 7]) == actual[1].flatten()).all())
        self.assertTrue((np.array([8, 9, 12, 13]) == actual[2].flatten()).all())
        self.assertTrue((np.array([10, 11, 14, 15]) == actual[3].flatten()).all())

    def test_extract_simple_patches_2d_color(self):
        image = np.arange(48).reshape((4, 4, 3))

        actual = extract_simple_patches_2d(image, (2, 2,))
        self.assertEqual((4, 2, 2, 3), actual.shape)

        self.assertTrue((np.array([0, 1, 2, 3, 4, 5, 12, 13, 14, 15, 16, 17]) == actual[0].flatten()).all())
        self.assertTrue((np.array([6, 7, 8, 9, 10, 11, 18, 19, 20, 21, 22, 23]) == actual[1].flatten()).all())
        self.assertTrue((np.array([24, 25, 26, 27, 28, 29, 36, 37, 38, 39, 40, 41]) == actual[2].flatten()).all())
        self.assertTrue((np.array([30, 31, 32, 33, 34, 35, 42, 43, 44, 45, 46, 47]) == actual[3].flatten()).all())

    def test_reconstruct_from_simple_patches_2d_gray(self):
        patches = np.stack((
            np.array([[0, 1], [4, 5]]),
            np.array([[2, 3], [6, 7]]),
            np.array([[8, 9], [12, 13]]),
            np.array([[10, 11], [14, 15]]),
        ))
        actual = reconstruct_from_simple_patches_2d(patches, (4, 4))
        self.assertTrue((np.arange(16).reshape((4, 4)) == actual).all())

    def test_reconstruct_from_simple_patches_2d_color(self):
        patches = np.stack((
            np.array([[[0, 1, 2], [3, 4, 5]], [[12, 13, 14], [15, 16, 17]]]),
            np.array([[[6, 7, 8], [9, 10, 11]], [[18, 19, 20], [21, 22, 23]]]),
            np.array([[[24, 25, 26], [27, 28, 29]], [[36, 37, 38], [39, 40, 41]]]),
            np.array([[[30, 31, 32], [33, 34, 35]], [[42, 43, 44], [45, 46, 47]]]),
        ))
        actual = reconstruct_from_simple_patches_2d(patches, (4, 4, 3))
        self.assertTrue((np.arange(48).reshape((4, 4, 3)) == actual).all())
